#include <stdio.h>
#include <math.h>
#include <nag.h>
#include <nag_stdlib.h>
#include <nagg01.h>
#include <nagg02.h>
#include <nagg12.h>
#include <nagf07.h>
#include <nagf16.h>
#include <nagx02.h>
#include <nagx04.h>

void forward_selection(Integer n, Integer m, const double z[], Integer sz[], 
                       const double t[], const Integer ic[], double pa, 
                       Integer *ip, double *dev, double *b[], double *se[], 
                       double *sc[], double *cov[], double *res[], Integer *nd,
                       double *tp[], double *sur[], Integer *info, FILE *fp);
void backward_selection(Integer n, Integer m, const double z[], Integer sz[], 
                        const double t[], const Integer ic[], double pd, 
                        Integer *ip, double *dev, double *b[], double *se[], 
                        double *sc[], double *cov[], double *res[], 
                        Integer *nd, double *tp[], double *sur[], 
                        Integer *info, FILE *fp);
void stepwise_selection(Integer n, Integer m, const double z[], Integer sz[],
                        const double t[], const Integer ic[], double pa, 
                        double pd, Integer *ip, double *dev, double *b[],
                        double *se[], double *sc[], double *cov[], 
                        double *res[], Integer *nd, double *tp[], 
                        double *sur[], Integer info[], FILE *fp);
double score_test_statistic(Integer ip, const double sc[], const double cov[]);
double wald_test_statistic(Integer ip, const double b[], const double cov[]);
void fit_coxph(Integer n, Integer m, const double z[], 
               const Integer sz[], Integer ip, const double t[], 
               const Integer ic[], double *dev, double *b[], double *se[], 
               double *sc[], double *cov[], double *res[], Integer *nd, 
               double *tp[], double *sur[], NagError *fail, Integer maxit); 
void estimate_initial_b(Integer n, Integer m, const double z[], 
                        const Integer sz[], Integer ip, const double t[], 
                        const Integer ic[],double *ib[]);
void select_var_to_add(Integer n, Integer m, const double z[], 
                       const Integer sz[], Integer *ip, const double t[], 
                       const Integer ic[], double pa, Integer csz[], 
                       double ib[], Integer *addvar, Integer info[], FILE *fp);
void select_var_to_drop(Integer m, double pd, Integer csz[], Integer *ip, 
                        double b[], double se[], Integer *dropvar, 
                        Integer info[], FILE *fp);

#define ENUMLEN 30

#define Z(I, J) z[((I))*m +(J)]

int main(void) {
  /* example program to demonstrate one possible way of implementing forward
     backwards and stepwise variable selection using g12bac */

  /* Integer scalar and array declarations */
  Integer i, m, n, nd, v, ip, j;
  Integer exit_status = 0;
  Integer *ic = 0, *sz = 0;
  Integer info[4];

  /* NAG structures and types */
  Nag_Boolean monit, wacc;
  NagError fail;

  /* double scalar and array declarations */
  double adj, dev, df, hazard, hazard_lci, hazard_uci, p, pa, pd, w;
  double *b = 0, *cov = 0, *res = 0, *sc = 0, *se = 0,
    *sur = 0, *t = 0, *tp = 0, *z = 0;

  /* character scalar and array declarations */
  char method, cmonit[ENUMLEN];

  /* other decalarations */
  FILE *fp = stdout;

  /* initialise the error structure */
  INIT_FAIL(fail);

  /* skip heading in data file */
  scanf("%*[^\n] ");

  /* read in the type of variable selection to use: 
     F = forwards, B = backwards and S = stepwise */
  scanf("%c%*[^\n] ",&method);

  /* read in the parameters that control the fitting process */
  switch (method) 
    {
    case 'f':
    case 'F':
      scanf("%lf%*[^\n] ",&pa);
      break;
      
    case 'b':
    case 'B':
      scanf("%lf%*[^\n] ",&pd);
      break;
      
    case 's':
    case 'S':
      scanf("%lf%lf%*[^\n] ",&pa,&pd);
      break;
      
    default:
      printf("Unknown variable selection method, method = %c\n", method);
      exit_status = -1;
      goto END;
    }

  /* read in monitoring flag */
  scanf("%s%*[^\n] ", cmonit);
  monit = (Nag_Boolean) nag_enum_name_to_value(cmonit);

  /* read in problem size */
  scanf("%"NAG_IFMT"%"NAG_IFMT"%*[^\n] ",&n, &m);

  z = NAG_ALLOC(n*m, double);
  assert(z);
  sz = NAG_ALLOC(m, Integer);
  assert(sz);
  t = NAG_ALLOC(n, double);
  assert(t);
  ic = NAG_ALLOC(n, Integer);
  assert(ic);
  
  /* read in the variable indication, this indicates which variables to include
     in the selection process, with SZ(i) = 1 indicating that the i'th variable
     is to be considered in the process and SZ(i) = 0 indicating that it is not
     considered */  
  for (i = ip = 0; i < m; i++) 
    {
      scanf("%"NAG_IFMT"",&sz[i]);
      if (sz[i] >= 1) ip++;
    }
  scanf("%*[^\n] ");

  /* read in the data */
  for (i = 0; i < n; i++) 
    {
      scanf("%lf",&t[i]);
      for (j = 0; j < m; j++)
        {
          scanf("%lf",&Z(i,j));
        }
      scanf("%"NAG_IFMT"%*[^\n] ",&ic[i]);
    }

  if (monit) 
    {
      fprintf(fp," Number of observations          :  %"NAG_IFMT"\n",n);
      fprintf(fp," Total no. of covariates         :  %"NAG_IFMT"\n", m);
      fprintf(fp," Number of covariates considered :  %"NAG_IFMT"\n", ip);
      fprintf(fp,"\n");
    }

  /* fit the Cox proportional hazards model using the specified variable 
     selection process */
  switch(method)
    {
    case 'f':
    case 'F':
      forward_selection(n,m,z,sz,t,ic,pa,&ip,&dev,&b,&se,&sc,&cov,&res,&nd,
                        &tp,&sur,info,fp);
      break;

    case 'b':
    case 'B':
      backward_selection(n,m,z,sz,t,ic,pd,&ip,&dev,&b,&se,&sc,&cov,&res,&nd,
                         &tp,&sur,info,fp);
      break;

    case 's':
    case 'S':
      stepwise_selection(n,m,z,sz,t,ic,pa,pd,&ip,&dev,&b,&se,&sc,&cov,&res,&nd,
                         &tp,&sur,info,fp);
      break;
    }

  /* check the error exit from the selection routine */
  if (info[0]!=NE_NOERROR)
    {
      if (info[0]==NW_POTENTIAL_PROBLEM)
        printf("Potential problem occurred in the selection routine\n");
      else
        printf("Fatal problem occurred in the selection routine\n");

      printf("INFO = ");
      for (i = 0; i < 4; i++)
        {
          printf(" %s", x04ndc(info[i]));
        }
      printf("\n");
      exit_status = -1;
      if (info[0]!=NW_POTENTIAL_PROBLEM) goto END;
    }

  /* display the information for the final model */
  if (ip==0) 
    {
      printf("\nNo covariates were included in the model, so there is no ");
      printf("model information to display\n");
    }
  else
    {
      printf("\n");
      printf("                            Parameter     Standard           ");
      printf("                      Hazard       95%% CI       95%% CI\n");
      printf("    Covariate        DF      Estimate        Error   Chi-Square");
      printf("   Pr > ChiSq        Ratio        Lower        Upper\n");
      printf("  ");
      for (i = 0; i < 113; i++)
        {
          printf("-");
        }
      printf("\n");

      i = -1;
      wacc = Nag_FALSE;
      for (v = 0; v < m; v++)
        {
          if (sz[v]==1)
            {
              i++;
              /* Wald test statistic and degrees of freedom as in the backwards
                 selection method, this again assumes that the covariates are 
                 continuous or binary (i.e. they have a single degree freedom).
                 If they do not then the wald_test_statistic routine would
                 need to be called and the degrees of freedom changed */
              w = (b[i]/se[i]);
              w*=w;
              df = 1.0;

              /* get the associated p-value */
              p = g01ecc(Nag_UpperTail,w,df,&fail);
              if (fail.code!=NE_NOERROR)
                {
                  if (fail.code!=NE_ALG_NOT_CONV) 
                    {
                      /* NE_ALG_NOT_CONV is a warning about accuracy, but it 
                         still returns a result so we will use it, otherwise 
                         call the routine again to trigger any messages and
                         terminate */
                      printf("Error from g01ecc.\n%s\n",fail.message);
                      exit_status = 1;
                      goto END;                      
                    }
                  else 
                    {
                      /* possible lack of accuracy */
                      wacc = Nag_TRUE;
                    }
                }

              /* get the width of the 95% confidence interval */
              adj = g01fac(Nag_TwoTailConfid,0.95,&fail)*se[i];

              /* calculate the harzard ratio estimate and associated confidence
                 interval */
              hazard = exp(b[i]);
              hazard_lci = exp(b[i]-adj);
              hazard_uci = exp(b[i]+adj);
              
              printf(" %10"NAG_IFMT"  %10.0f    %10.4f   %10.4f   %10.4f   ",
                     v+1, df, b[i], se[i], w);
              printf("%10.4f   %10.4f   %10.4f   %10.4f\n", p, hazard, 
                     hazard_lci, hazard_uci);
            }
        }
      if (wacc) 
        {
          printf("Warning: Some p-values have been flagged as having ");
          printf("accuracy issues, displayed values");
          printf(" should be a good approximation");
        }
      printf("\n");

      /* as we have assumed that all covariates are binary or continuous we 
         have not actually used the WALD_TEST_STATISTIC routine in this 
         example, so perform a global null hypothesis test of 
         H0: B = 0, vs H1: B != 0 */

      /* get the test statistic and degrees of freedom, based on the final 
         selected model (where B and COV come from the call to fit_coxph 
         above) */
      w = wald_test_statistic(ip,b,cov);
      df = (double) ip;

      /* get the associated p-value */
      p = g01ecc(Nag_UpperTail,w,df,&fail);
      if (fail.code!=NE_NOERROR)
        {
          if (fail.code!=NE_ALG_NOT_CONV) 
            {
              /* NE_ALG_NOT_CONV is a warning about accuracy, but it still 
                 returns a result so we will use it, otherwise call the 
                 routine again to trigger any messages and terminate */
              printf("Error from g01ecc.\n%s\n",fail.message);
              exit_status = 1;
              goto END;                      
            }
          else 
            {
              /* possible lack of accuracy */
              wacc = Nag_TRUE;
            }
        }

      /* display the test */
      printf(" Global Null Hypothesis B = 0\n");
      printf(" ----------------------------\n");
      printf("      Wald\n");
      printf("    Statistic         DF       p-value\n");
      printf("   -----------------------------------\n");
      printf(" %10.4f  %10.0f    %10.4f\n", w, df, p);
      if (wacc) 
        {
          printf(" Warning: The p-value has been flagged as having accuracy");
          printf("issues, displayed value");
          printf(" should be a good approximation\n");
        }
    }

 END:
  if (b) NAG_FREE(b);
  if (cov) NAG_FREE(cov);
  if (res) NAG_FREE(res);
  if (sc) NAG_FREE(sc);
  if (se) NAG_FREE(se);
  if (sur) NAG_FREE(sur);
  if (t) NAG_FREE(t);
  if (tp) NAG_FREE(tp);
  if (z) NAG_FREE(z);
  if (ic) NAG_FREE(ic);
  if (sz) NAG_FREE(sz);

  return(exit_status);
}

void forward_selection(Integer n, Integer m, const double z[], Integer sz[], 
                       const double t[], const Integer ic[], double pa, 
                       Integer *ip, double *dev, double *b[], double *se[], 
                       double *sc[], double *cov[], double *res[], Integer *nd,
                       double *tp[], double *sur[], Integer info[], FILE *fp)
{
  /* Fits a Cox proportional hazards model using a forwards selection process 
     to automatically choose which model to fit.
  
     The selection process uses the Score test statistic to determine which 
     covariate (if any) is added to the model at each iteration 
  
     Input: 
     n, m, z, t and 
     ic   - As per g12bac 
     pa   - Cut-off used when deciding whether a covariate should be included 
            in the model. If the p-value associated with the largest Score 
            test statistic <= pa then the associated covariate is included 
            in the model 
     fp   - If non-null then monitoring information is sent to FP, otherwise no
            information is printed 
  
     Input / Output:
     sz   - On entry: indicates which subset of covariates are to be included 
            in the model selection process: 
              sz[j+1] >= 1 indicates that the jth covariate is included in the 
                         process 
              sz[j+1] = 0  indicates that the jth covariate is ignored 
            On exit: indicates which subset of covariates are in the final model 
              sz[j+1] >= 1 indicates that the jth covariate is included in
                           the model 
              sz[j+1] = 0  indicates that the jth covariate is not included in 
                           the model 
     ip   - On entry, the number of covariates included in the model 
            On exit, the (updated) number of covariates included in the model

     Output: 
     dev,b,se,sc,cov,res,bd,tp and 
     sur  - As per g12bac. These output variables are for the final model as 
            defined by the values of sz on exit. 
     info - Warning / error information, info[4]
            info[0] = NE_NOERROR if all is OK
                    = NW_POTENTIAL_PROBLEM if a potential problem occurred, but
                      output still returned
                    = NE_BAD_PARAM indicates that an error occurred somewhere
                      in the routine and no output could be returned
                      Information on any errors or warnings is returned in 
                      the remaining elements of INFO
           info[1]  = value of fail.code from fit_coxph 
           info[2]  = value of fail.code from g01ecc
           info[3]  = not used
  */
  
  /* Integer scalar and array declarations */
  Integer addvar, iter, max_iter, i, j;
  Integer *csz = 0;
 
  /* NAG structures and types */
  NagError fail;

  /* double scalar and array declarations */
  double *ib = 0;

  for (i = 0; i < 4; i++) info[i] = NE_NOERROR;

  /* get the maximum number of iterations, which is the number of covariates 
     that will be considered as defined by sz[i] >= 1 */
  for (i = max_iter = 0; i < m; i++)
    if (sz[i] >= 1) max_iter++;

  /* array to hold the covariates included in the current model */
  csz = NAG_ALLOC(m, Integer);
  assert(csz);

  /* array to hold the parameter estimates for the current model unlike the 
     array b, in g12bac which only has entries for those covariates in the 
     current model, ib will have entries for all m covariates and 
     ib[i] = 0.0 if csz[i] = 0 (i.e. ib[i] will take a value of zero if 
     covariate i+1 is not in the model) */
  ib = NAG_ALLOC(m, double);
  assert(ib);

  /* start with the null model */
  for (i = 0; i < m; i++)
    {
      csz[i] = 0;
      ib[i] = 0.0;
    }
  *ip = 0;
  
  /* loop over the number of possible models to consider */
  for (iter = 1; iter <= max_iter; iter++)
    {
      if (fp)
        {
          /* display a heading if we are outputing monitoring information */
          fprintf(fp,"\n Iteration %"NAG_IFMT"\n", iter);
          fprintf(fp," -----------\n");
        }

      /* select the covariate to add */
      select_var_to_add(n,m,z,sz,ip,t,ic,pa,csz,ib,&addvar,info,fp);
      if (addvar < 1)
        {
          if (fp)
            {
              fprintf(fp,"No covariates added to the model at iteration");
              fprintf(fp," %"NAG_IFMT"\n", iter);
              fprintf(fp,"Terminating selection process\n");
            }
          break;
        }

      if (fp) 
        {
          fprintf(fp,"Covariate %"NAG_IFMT" added to the model at",addvar);
          fprintf(fp," iteration %"NAG_IFMT"\n", iter);
        }

      /* fit the current model (use default values for all the optional 
         arguments) */
      fit_coxph(n,m,z,csz,*ip,t,ic,dev,b,se,sc,cov,res,nd,tp,sur,&fail,-1);
      if (fail.code != NE_NOERROR)
        {
          info[1] = fail.code;
          if (fail.code != NE_G12BA_DEV && fail.code != NE_G12BA_CONV)
            {
              info[0] = NE_BAD_PARAM;
              break;
            }
          info[0] = NW_POTENTIAL_PROBLEM;
          /* fail.code = NE_G12BA_DEV and NE_G12BA_CONV indicate possible 
             non-covergence, but the full results will have been returned, 
             so continue */
        }

      /* copy the parameter estimates from the current model into ib, so ib[i]
         will take a value from b if csz[i] >= 1 and remain unchanged otherwise
         (which means it will be 0). If csz[i] >= 1 and it is the jth value of 
         1 in csz then ib[i] = b[j] */
      for (i = j = 0; i < m; i++)
        {
          if (csz[i]>=1) {
            ib[i] = (*b)[j];
            j++;
          }
        }
    }

  /* update the information in SZ to the last model */
  for (i = 0; i < m; i++) sz[i] = csz[i];
      
  /* deallocate the local allocatable arrays */
  if (csz) NAG_FREE(csz);
  if (ib) NAG_FREE(ib);
}

void backward_selection(Integer n, Integer m, const double z[], Integer sz[], 
                        const double t[], const Integer ic[], double pd, 
                        Integer *ip, double *dev, double *b[], double *se[], 
                        double *sc[], double *cov[], double *res[], Integer *nd,
                        double *tp[], double *sur[], Integer info[], FILE *fp)
{
  /* Fits a Cox proportional hazards model using a backwards selection process 
     to automatically choose which model to fit.
  
     The selection process uses the Wald test statistic to determine which 
     covariate (if any) is dropped from the model at each iteration 
  
     Input: 
     n, m, z, t and 
     ic   - As per g12bac 
     pd   - Cut-off used when deciding whether a covariate should be dropped 
            from the model. If the p-value associated with the smallest Score 
            test statistic > pd then the associated covariate is dropped
            from the model 
     fp   - If non-null then monitoring information is sent to fp, otherwise no
            information is printed 
  
     Input / Output:
     sz   - On entry: indicates which subset of covariates are to be included 
            in the model selection process: 
              sz[j+1] >= 1 indicates that the jth covariate is included in the 
                         process 
              sz[j+1] = 0  indicates that the jth covariate is ignored 
            On exit: indicates which subset of covariates are in the final model 
              sz[j+1] >= 1 indicates that the jth covariate is included in
                           the model 
              sz[j+1] = 0  indicates that the jth covariate is not included in 
                           the model 
     ip   - On entry, the number of covariates included in the model 
            On exit, the (updated) number of covariates included in the model

     Output: 
     dev,b,se,sc,cov,res,bd,tp and 
     sur  - As per g12bac. These output variables are for the final model as 
            defined by the values of sz on exit. 
     info - Warning / error information, info[4]
            info[0] = NE_NOERROR if all is OK
                    = NW_POTENTIAL_PROBLEM if a potential problem occurred, but
                      output still returned
                    = NE_BAD_PARAM indicates that an error occurred somewhere
                      in the routine and no output could be returned
                      Information on any errors or warnings is returned in 
                      the remaining elements of INFO
           info[1]  = value of fail.code from fit_coxph 
           info[2]  = value of fail.code from g01ecc
           info[3]  = NE_VAR_EQ_ZERO if can't calculate Wald test
                      statistic due to a zero variance, NE_NOERROR otherwise
  */

  /* Integer scalar and array declarations */
  Integer dropvar, iter, max_iter, i;
  Integer *csz = 0;

  /* NAG structures and types */
  NagError fail;

  for (i = 0; i < 4; i++) info[i] = NE_NOERROR;

  /* get the maximum number of iterations, which is the number of covariates */
  /* that will be considered as defined by sz[i] >= 1 */
  for (i = max_iter = 0; i < m; i++)
    if (sz[i] >= 1) max_iter++;

  /* array to hold the covariates included in the current model */
  csz = NAG_ALLOC(m, Integer);
  assert(csz);

  /* start with the full model */
  for (i = 0; i < m; i++) csz[i] = 1;
  *ip = m;

  /* loop over the number of possible models to consider */
  for (iter = 1; iter <= max_iter; iter++)
    {

      /* fit the current model (using default values for all the optional 
         parameters) */
      fit_coxph(n,m,z,csz,*ip,t,ic,dev,b,se,sc,cov,res,nd,tp,sur,&fail,-1);
      if (fail.code != NE_NOERROR)
        {
          info[1] = fail.code;
          if (fail.code != NE_G12BA_DEV && fail.code != NE_G12BA_CONV)
            {
              info[0] = NE_BAD_PARAM;
              break;
            }
          info[0] = NW_POTENTIAL_PROBLEM;
          /* fail.code = NE_G12BA_DEV and NE_G12BA_CONV indicate possible 
             non-covergence, but the full results will have been returned, 
             so continue */
        }

      if (fp)
        {
          /* display a heading if we are outputing monitoring information */
          fprintf(fp,"\n Iteration %"NAG_IFMT"\n", iter);
          fprintf(fp," -----------\n");
        }

      /* select the covariate to drop */
      select_var_to_drop(m,pd,csz,ip,*b,*se,&dropvar,info,fp);
      if (dropvar<1)
        {
          if (fp)
            {
              fprintf(fp,"No covariates dropped from the model at ");
              fprintf(fp,"iteration %"NAG_IFMT"\n", iter);
              fprintf(fp,"Terminating selection process\n");
            }
          break;
        }

      if (fp) 
        {
          fprintf(fp,"Covariate %"NAG_IFMT" dropped from the model at ", 
                  dropvar);
          fprintf(fp,"iteration %"NAG_IFMT"\n", iter);
        }
    }

  /* update the information in SZ to the last model */
  for (i = 0; i < m; i++) sz[i] = csz[i];

  /* deallocate the local allocatable arrays */
  if (csz) NAG_FREE(csz);
}

void stepwise_selection(Integer n, Integer m, const double z[], Integer sz[], 
                        const double t[], const Integer ic[], double pa, 
                        double pd, Integer *ip, double *dev, double *b[], 
                        double *se[], double *sc[], double *cov[], 
                        double *res[], Integer *nd, double *tp[], double *sur[],
                        Integer info[], FILE *fp)
{
  /* Fits a Cox proportional hazards model using a stepwise selection process 
     to automatically choose which model to fit.

     The selection process uses the Score test statistic to determine which
     covariate (if any) is added to the model and the Wald
     test statistic to determine which covariate (if any) is dropped from
     the model at each iteration
  
     Input: 
     n, m, z, t and 
     ic   - As per g12bac 
     pa   - Cut-off used when deciding whether a covariate should be included 
            in the model. If the p-value associated with the largest Score 
            test statistic <= pa then the associated covariate is included 
            in the model 
     pd   - Cut-off used when deciding whether a covariate should be dropped 
            from the model. If the p-value associated with the smallest Score 
            test statistic > pd then the associated covariate is dropped
            from the model 
     fp   - If non-null then monitoring information is sent to FP, otherwise no
            information is printed 
  
     Input / Output:
     sz   - On entry: indicates which subset of covariates are to be included 
            in the model selection process: 
              sz[j+1] >= 1 indicates that the jth covariate is included in the 
                         process 
              sz[j+1] = 0  indicates that the jth covariate is ignored 
            On exit: indicates which subset of covariates are in the final model 
              sz[j+1] >= 1 indicates that the jth covariate is included in
                           the model 
              sz[j+1] = 0  indicates that the jth covariate is not included in 
                           the model 
     ip   - On entry, the number of covariates included in the model 
            On exit, the (updated) number of covariates included in the model

     Output: 
     dev,b,se,sc,cov,res,bd,tp and 
     sur  - As per g12bac. These output variables are for the final model as 
            defined by the values of sz on exit. 
     info - Warning / error information, info[4]
            info[0] = NE_NOERROR if all is OK
                    = NW_POTENTIAL_PROBLEM if a potential problem occurred, but
                      output still returned
                    = NE_BAD_PARAM indicates that an error occurred somewhere
                      in the routine and no output could be returned
                      Information on any errors or warnings is returned in 
                      the remaining elements of INFO
           info[1]  = value of fail.code from fit_coxph 
           info[2]  = value of fail.code from g01ecc
           info[3]  = NE_VAR_EQ_ZERO if can't calculate Wald test
                      statistic due to a zero variance, NE_NOERROR otherwise
  */

  /* Integer scalar and array declarations */
  Integer addvar, dropvar, iter, max_iter, i, j;
  Integer *csz = 0;
 
  /* NAG structures and types */
  NagError fail;

  /* double scalar and array declarations */
  double *ib = 0;

  for (i = 0; i < 4; i++) info[i] = NE_NOERROR;

  /* get the maximum number of iterations, which is the number of covariates
     that will be considered as defined by sz[i] >= 1 */
  for (i = max_iter = 0; i < m; i++)
    if (sz[i] >= 1) max_iter++;

  /* array to hold the covariates included in the current model */
  csz = NAG_ALLOC(m, Integer);
  assert(csz);

  /* array to hold the parameter estimates for the current model unlike the 
     array b, in g12bac which only has entries for those covariates in the 
     current model, ib will have entries for all m covariates and ib[i] = 0.0
     if csz[i] = 0 (i.e. ib[i] will take a value of zero if covariate i+1 is 
     not in the model) */
  ib = NAG_ALLOC(m, double);
  assert(ib);

  /* start with the null model */
  for (i = 0; i < m; i++)
    {
      csz[i] = 0;
      ib[i] = 0.0;
    }
  *ip = 0;
  
  /* loop over the number of possible models to consider */
  for (iter = 1; iter <= max_iter; iter++)
    {
      if (fp)
        {
          /* display a heading if we are outputing monitoring information */
          fprintf(fp,"\n Iteration %"NAG_IFMT"\n", iter);
          fprintf(fp," -----------\n");
          fprintf(fp,"\n Forwards Step\n");
          fprintf(fp," -------------\n");
        }

      /* select the covariate to add */
      select_var_to_add(n,m,z,sz,ip,t,ic,pa,csz,ib,&addvar,info,fp);
      if (addvar < 1)
        {
          if (fp)
            {
              fprintf(fp,"No covariates added to the model at iteration");
              fprintf(fp," %"NAG_IFMT"\n", iter);
              fprintf(fp,"Terminating selection process\n");
            }
          break;
        }

      if (fp) 
        {
          fprintf(fp,"Covariate %"NAG_IFMT" added to the model at iteration", 
                  addvar);
          fprintf(fp," %"NAG_IFMT"\n", iter);
        }

      /* fit the current model (use default values for all the optional 
         arguments) */
      fit_coxph(n,m,z,csz,*ip,t,ic,dev,b,se,sc,cov,res,nd,tp,sur,&fail,-1);
      if (fail.code != NE_NOERROR)
        {
          info[1] = fail.code;
          if (fail.code != NE_G12BA_DEV && fail.code != NE_G12BA_CONV)
            {
              info[0] = NE_BAD_PARAM;
              break;
            }
          info[0] = NW_POTENTIAL_PROBLEM;
          /* fail.code = NE_G12BA_DEV and NE_G12BA_CONV indicate possible 
             non-covergence, but the full results will have been returned, 
             so continue */
        }

      /* copy the parameter estimates from the current model into ib, so ib[i]
         will take a value from b if csz[i] >= 1 and remain unchanged otherwise 
         (which means it will be 0). If csz[i] >= 1 and it is the jth value of
         1 in csz then ib[i] = b[j] */
      for (i = j = 0; i < m; i++)
        {
          if (csz[i]>=1) {
            ib[i] = (*b)[j];
            j++;
          }
        }

      if (fp)
        {
          fprintf(fp,"\n Backwards Step\n");
          fprintf(fp," --------------\n");
        }

      /* select the covariate to drop */
      select_var_to_drop(m,pd,csz,ip,*b,*se,&dropvar,info,fp);
      if (fp)
        {
          if (dropvar<1)
            {
              fprintf(fp,"No covariates dropped from the model at iteration ");
              fprintf(fp,"%"NAG_IFMT"\n", iter);
            }
          else
            {
              fprintf(fp,"Covariate %"NAG_IFMT" dropped from the ",dropvar);
              fprintf(fp,"model at iteration %"NAG_IFMT"\n", iter); 
            }
        }

      /* check that we havn't dropped the same variable as we just added */
      if (dropvar==addvar)
        {
          if (fp)
            {
              fprintf(fp,"The same covariate was added and dropped at this");
              fprintf(fp," iteration\nTerminating selection process\n");
            }

          /* we have not called fit_coxph on the current model, the last call 
             was with dropvar included in the model, so if there are some
             variables in this model we need to call fit_coxph again */
          if (*ip > 0)
            {
              fit_coxph(n,m,z,csz,*ip,t,ic,dev,b,se,sc,cov,res,nd,tp,
                        sur,&fail,-1);
              if (fail.code != NE_NOERROR)
                {
                  info[1] = fail.code;
                  if (fail.code != NE_G12BA_DEV && fail.code != NE_G12BA_CONV)
                    {
                      info[0] = NE_BAD_PARAM;
                      break;
                    }
                  info[0] = NW_POTENTIAL_PROBLEM;
                }
            }
          break;
        }
    }
  
  /* update the information in SZ to the last model */
  for (i = 0; i < m; i++) sz[i] = csz[i];
      
  /* deallocate the local allocatable arrays */
  if (csz) NAG_FREE(csz);
  if (ib) NAG_FREE(ib);
}

double score_test_statistic(Integer ip, const double sc[], const double cov[])
{
  /* Calculate and return the Score test statistic, S

     Input:                                                              
     ip     - Number of covariates in the current model
     sc     - Value of the score function as returned by g12bac
     cov    - Covariance matrix, in packed format, as returned by g12bac
  */

  /* Integer scalar and array declarations */
  Integer i;

  /* double scalar and array declarations */
  double s;
  double *covsc = 0;

  /* allocate some workspace */
  covsc = NAG_ALLOC(ip, double);
  assert(covsc);

  /* f16pec: calculate cov * sc,
     we using the default error structure, which will terminate if an error 
     occurs as we should only ever be supplying valid input arguments, so 
     routine should not fail */
  nag_dspmv(Nag_ColMajor,Nag_Upper,ip,1.0,cov,sc,1,0.0,covsc,1,NAGERR_DEFAULT);

  /* calculate transpose(sc) * cov * sc, which gives the Score test statistic */
  for (i = 0, s = 0.0; i < ip; i++) s+= sc[i] * covsc[i];
  
  if (covsc) NAG_FREE(covsc);

  return s;
}

double wald_test_statistic(Integer ip, const double b[], const double cov[])
{
  /* Calculate and return the Wald test statistic, W

     Input:
     ip     - Number of covariates in the current model
     b      - Parameter estimates as returned by g12abac
     cov    - Covariance matrix, in packed format, as returned by g12bac
  */

  /* Integer scalar and array declarations */
  Integer i, lcov;
  
  /* double scalar and array declarations */
  double w;
  double *ccov = 0, *cb = 0;

  /* number of elements in cov */
  lcov = ip*(ip+1)/2;

  /* copy cov and b so they don't get overwritten */

  /* copy cov */
  ccov = NAG_ALLOC(lcov, double);
  assert(cov);
  for (i = 0; i < lcov; i++) ccov[i] = cov[i];

  /* f07gdc: factorise cov so that cov = transpose(U) * U, where ccov = cov on
     entry and U on exit. As of CL23, the C library does not have a pivotted 
     Cholesky factorisation so we are using the standard Cholesky. From CL24 
     onwards it is likely to be better to use the pivotted Cholesky.
     We are using the default error structure, NAGERR_DEFAULT, which terminates
     if an error occurs. We are only ever supplying valid values to nag_dpptrf.
     Therefore the only possible error exit is NE_POS_DEF, if ccov is not 
     positive definite. As ccov is a covariance matrix it is possible that it
     is semi-positive rather than positive however, we have not coded to 
     account for this case as it is likely to be rare. This would not be an 
     issue if a pivotted cholesky factorisation was used. */
  nag_dpptrf(Nag_ColMajor,Nag_Upper,ip,ccov,NAGERR_DEFAULT);

  /* copy b */
  cb = NAG_ALLOC(ip, double);
  assert(cb);
  for (i = 0; i < ip; i++) cb[i] = b[i];

  /* f16plc: solve ccov * x = cb for x, putting the result in cb again we are
     using the default error structure. Only possible error exits from this
     routine are for invalid input arguments which should never happen here */
  nag_dtpsv(Nag_ColMajor,Nag_Upper,Nag_Trans,Nag_NonUnitDiag,ip,1.0,ccov,cb,
            1,NAGERR_DEFAULT);

  /* calculate cb.cb = transpose(b) * inverse(cov) * b, which gives the Wald
     test statistic */
  for (i = 0, w = 0.0; i < ip; i++) w+= cb[i] * cb[i];

  if (ccov) NAG_FREE(ccov);
  if (cb) NAG_FREE(cb);

  return w;
}

void fit_coxph(Integer n, Integer m, const double z[], const Integer sz[], 
               Integer ip, const double t[], const Integer ic[], double *dev, 
               double *b[], double *se[], double *sc[], double *cov[], 
               double *res[], Integer *nd, double *tp[], double *sur[], 
               NagError *fail, Integer maxit)
{
  /* Fit a Cox proporational hazards model using g12bac this routine is less
     flexible than g12bac as it simplifies the g12bac interface somewhat, by
     using fixed values for tolerances etc and not allowing an offset or
     weights 

     initial values are estimated using estimate_initial_b, unless maxit = 0 

     Input: 
     n,m,z,sz,ip,t and 
     IC       - Are as per g12bac 
     maxit    - The maximum number of iterations, if maxit < 0 then a default 
                value of 1000 iterations is used 
     
     Input / Output: 
     b        - On entry, If maxit = 0, then the parameter estimates are taken
                from (*b) otherwise, the initial parameter estimates are 
                estimated from the data using estimate_initial_b
                On exit, the address of a vector holding the parameter estimates 
     Output: 
     dev,se,sc,cov,res,nd,tp and 
     sur      - Pointers to scalars or vectors holding the various output 
                values returned by g12bac. Output arrays are locally allocated.
                So, for example, sc will point to an array holding the output
                vector of the same name from g12bac     
     faile    - Error structure as per g12bac 
  */

  /* Integer scalar and array declarations */
  Integer tdz, tdsur, iprint, mmaxit, ndmax, ns;
  Integer *isi = 0;

  /* double scalar and array declarations */
  double tol;
  double *omega = 0;

  /* character scalars and array declarations */
  char *outfile = 0;

  /* initialise the error structure */
  INIT_FAIL(*fail);

  /* we are not allowing strata */
  ns = 0;
  isi = 0;

  /* we are assuming that all the arrays are the minimum size required */
  tdz = m;
  tdsur = 1;

  /* no offset */
  omega = 0;
 
  /* no monitoring information */
  iprint = 0;
  outfile = 0;

  /* set the control parameters to default values */
  tol = 10.0 * X02AJC;
  mmaxit = (maxit < 0) ? 1000 : maxit;

  /* allocate the rest of the output arrays */
  ndmax = n;

  /* deallocate the various output arrays, if already allocated */
  if (*se) NAG_FREE(*se);
  if (*sc) NAG_FREE(*sc);
  if (*cov) NAG_FREE(*cov);
  if (*res) NAG_FREE(*res);
  if (*tp) NAG_FREE(*tp);
  if (*sur) NAG_FREE(*sur);

  /* allocate output arrays */
  (*se) = NAG_ALLOC(ip,double);
  assert(*se);
  (*sc) = NAG_ALLOC(ip,double);
  assert(*sc);
  (*cov) = NAG_ALLOC(ip*(ip+1)/2,double);
  assert(*cov);
  (*res) = NAG_ALLOC(n,double);
  assert(*res);
  (*tp) = NAG_ALLOC(ndmax,double);
  assert(*tp);
  (*sur) = NAG_ALLOC(ndmax*tdsur,double);
  assert(*sur);


  /* get initial values if required, by fitting an exponential model */
  if (mmaxit > 0 || !b) 
    {
      estimate_initial_b(n,m,z,sz,ip,t,ic,b);
    }

  /* fit Cox proportional hazards model */
  g12bac(n,m,ns,z,tdz,sz,ip,t,ic,omega,isi,dev,*b,*se,*sc,*cov,*res,nd,*tp,
         *sur,tdsur,ndmax,tol,mmaxit,iprint,outfile,fail);
  
  /* adjust the error exit from the NAG routine */
  if (maxit == 0 && (*fail).code == NE_G12BA_CONV) 
    {
      /* when we set MAXIT = 0 we know that g12bac will not converge, so ignore
         those warning messages */
      INIT_FAIL(*fail);
    }
} 

void estimate_initial_b(Integer n, Integer m, const double z[], 
                        const Integer sz[], Integer ip, const double t[],
                        const Integer ic[],double *ib[])
{
  /* estimate initial values for the parameter estimates of a Cox proportional 
     hazards model using an exponential model fitted via G02GCF 

     Input: 
     n,m,z,sz,ip,t and 
     ic    - As per g12bac 
     
     Output: 
     ib    - Initial values to pass to g12bac 

     ib need not be allocated on input, it is (re)allocated by this routine,
     if g02gcc fails for any reason then ib(i) = 0 for all i is returned 
  */

  /* Integer scalar and array declarations */
  Integer i, print_iter, irank, tdv, tdz, max_iter, tip;

  /* NAG structures and types */
  NagError fail;
  Nag_Link link;
  Nag_IncludeMean mean;

  /* double scalar and array declarations */
  double ex_power, dev, eps, tol, df;
  double *b = 0, *cov = 0, *se =0, *v = 0, *wt = 0, *y = 0, *offset = 0;

  /* character scalars and array declarations */
  char *outfile = 0;

  /* initialise the error structure */
  INIT_FAIL(fail);

  /* we are fitting a mean in the exponential model, so there is one more
     covariate (the mean) than that specified in ip */
  tip = ip + 1;

  /* assume arrays are of the minimum size */
  tdz = m;
  tdv = tip + 6;

  /* allocate output arrays for g02gcc */
  y = NAG_ALLOC(n,double);
  offset = NAG_ALLOC(n,double);
  b = NAG_ALLOC(tip,double);
  se = NAG_ALLOC(tip,double);
  cov = NAG_ALLOC(tip*(tip+1)/2,double);
  v = NAG_ALLOC(n*tdv,double);

  /* format data in the way g02gcc expects */
  for (i = 0; i < n; i++)
    {
      y[i] = 1.0 - (double) ic[i];
      offset[i] = log(t[i]);
    }

  /* unweighted, log link, with mean */
  link = Nag_Log;
  mean = Nag_MeanInclude;
  wt = 0;
  ex_power = 0.0;
  
  /* use default values for control parameters */
  tol = 0.0;
  max_iter = 0;
  eps = 0.0;

  /* no printing */
  print_iter = 0;
  outfile = 0;
  
  /* fit exponential model */
  g02gcc(link,mean,n,z,tdz,m,sz,tip,y,wt,offset,ex_power,&dev,&df,b,&irank,se,
         cov,v,tdv,tol,max_iter,print_iter,outfile,eps,&fail);

  /* allocate output array */
  if (*ib) NAG_FREE(*ib);
  (*ib) = NAG_ALLOC(ip,double);

  if (fail.code == NE_NOERROR || fail.code == NE_LSQ_ITER_NOT_CONV || 
      fail.code == NE_VALUE_AT_BOUNDARY_C || fail.code == NE_RANK_CHANGED)
    {
      /* use the estimates returned by g02gcc, but we need to move all */
      /* the parameter estimates down one so as to drop the parameter */
      /* estimate for the mean */
      for (i = 0; i < ip; i++) (*ib)[i] = b[i+1];
    }
  else
    {
      /* g02gcc has failed, so just use 0 as the initial values */
      for (i = 0; i < ip; i++) (*ib)[i] = 0.0;
    }

  /* deallocate the temporary arrays we no longer require */
  if (y) NAG_FREE(y);
  if (offset) NAG_FREE(offset);
  if (b) NAG_FREE(b);
  if (se) NAG_FREE(se);
  if (cov) NAG_FREE(cov);
  if (v)  NAG_FREE(v);
}

void select_var_to_add(Integer n, Integer m, const double z[], 
                       const Integer sz[], Integer *ip, const double t[],
                       const Integer ic[], double pa, Integer csz[], 
                       double ib[], Integer *addvar, Integer info[], FILE *fp)
{
  /* Select a covariate to add to a Cox proportional hazards model using the 
     Score statistic and its associated p-value

     Input: 
     n,m,z,t,ic,pa and 
     fp     - As per forward_selection 
     sz     - As described in forward_selection, on entry 
     csz    - On entry: indicates which subset of covariates are in the current
              model, with 
               csz[j+1] >= 1 indicating that the jth covariate is in the
                             current model 
               csz[j+1] = 0  indicating that the jth covariate is not in the 
                             current model 
              On exit: as per entry, but if addvar >= 1, then csz[addvar-1]  
              will have been set to 1 as well (i.e. on exit csz indicates 
              the covariates included in the new model, with covariate 
              indicated by addvar having been added) 
     ib     - Parameter estimates, based on the current model as described 
              by csz. ib is of length m and contains values for all covariates 
              If covariate j is not in the current model then ib[j+1] = 0.0 
              (this is a different structure to the parameter b in g12bac 
              which only contains values for the *ip covariates in the model)

     Input / Output:
     ip     - On entry, the number of covariates in the currrent model
              On exit, the (updated) number of covariates in the current model
     
     Output: 
     addvar - The covariate to add to the model, If addvar > 0 then 
              1 <= addvar <= m. If addvar < 1 then no covariates satisfied the
              criteria defined by pa and so no covariate was added to the model
     info   - As per forward_selection 
  */

  /* Integer scalar and array declarations */
  Integer cip, j, mj, nd, i, v;

  /* NAG structures and types */
  NagError fail;
  Nag_Boolean wacc;

  /* double scalar and array declarations */
  double dev, df, max_s, p, s;
  double *b = 0, *cov = 0, *res = 0, *sc = 0, *se = 0, *sur = 0, *tp = 0;

  /* initialise the error structure */
  INIT_FAIL(fail);

  /* initialise the variable holding the maximum score (max_s), the 
     associated covariate (mj) and the variable to add */
  max_s = 0.0;
  mj = 0;
  *addvar = -1;

  /* we will be considering a model with one additional covariate this assumes 
     that all the covariates are continous or binary (i.e. have a single degree
     of freedom). If this is not the case then more covariates might be added 
     at each step */
  cip = *ip + 1;

  /* work array to hold the parameter estimates */
  b = NAG_ALLOC(cip,double);
  assert(b);

  /* display column headers for monitoring information */
  if (fp) fprintf(fp,"   Covariate    Score\n");

  /* loop over all possible covariates */
  for (j = 0; j < m; j++)
    {
      /* only want to look at covariates included in the selection process
         (as defined by sz[j] >= 1) and those not in the current model
         (as defined by csz[j] = 0) */
      if (sz[j]>=1 && csz[j]==0)
        {

          /* add the covariate being considered to the model */
          csz[j] = 1;

          /* the order of the values in b must be the same as the order of 
             elements of csz >= 1 */
          for (i = v = 0; i < m; i++)
            {
              if (csz[i]>=1)
                {
                  b[v] = ib[i];
                  v++;
                }
            }

          /* calculate the score function and covariance matrix as we do not 
             want to calculate the parameter estimates but only the values of 
             the other output arguments, set maximum iterations to zero */
          fit_coxph(n,m,z,csz,cip,t,ic,&dev,&b,&se,&sc,&cov,&res,&nd,&tp,&sur,
                    &fail,0);
          if (fail.code != NE_NOERROR)
            {
              info[0] = NE_BAD_PARAM;
              info[1] = fail.code;
              goto END;
            }

          /* calculate the Score test statistic */
          s = score_test_statistic(cip,sc,cov);
          if (s > max_s)
            {
              max_s = s;
              mj = j;
            }

          if (fp)
            {
              /* display the individual scores */
              fprintf(fp,"   %5"NAG_IFMT"   %10.4f\n",j+1,s);
            }

          /* reset the list of covariates */
          csz[j] = 0;
        }
    }

  /* degrees of freedom. This assumes that the covariates are either 
     continuous or binary (i.e. they have a single degree of freedom). 
     If this is not the case then the calculation of the degrees of freedom 
     would need changing */
  df = 1.0;

  /* get the p-value for max_s */
  wacc = Nag_FALSE;
  p = g01ecc(Nag_UpperTail,max_s,df,&fail);
  if (fail.code!=NE_NOERROR)
    {
      info[2] = fail.code;
      if (fail.code!=NE_ALG_NOT_CONV)
        {
          info[0] = NE_BAD_PARAM;
          goto END;
        }
      wacc = Nag_TRUE;
      /* fail.code = NE_ALG_NOT_CONV is an accuracy warning, but the returned 
         value should be sufficient so continue */
    }

  if (fp)
    {
      /* print summary information for this iteration */
      fprintf(fp,"\n   Covariate %"NAG_IFMT" considered for addition\n",mj+1);
      fprintf(fp,"       Score   = %10.4f\n", max_s);
      fprintf(fp,"       p-value = %10.6f\n", p);
      if (wacc)
        {
          fprintf(fp,"          convergence failed in calculation of p-value\n");
          fprintf(fp,"          reported value should be a good approximation\n");
        }
      fprintf(fp,"       cut-off = %10.6f\n\n", pa);
    }

  /* check whether the p-value associated with the largest value of S meets the
     criteria defined by pa */
  if (p>pa)
    {
      *addvar = -1;
    }
  else
    {
      *addvar = mj + 1;

      /* add the covariate to the model */
      csz[mj] = 1;
      (*ip)++;
    }

 END:
  if (b) NAG_FREE(b);
  if (se) NAG_FREE(se);
  if (sc) NAG_FREE(sc);
  if (cov) NAG_FREE(cov);
  if (res) NAG_FREE(res);
  if (tp) NAG_FREE(tp);
  if (sur) NAG_FREE(sur);
}

void select_var_to_drop(Integer m, double pd, Integer csz[], Integer *ip, 
                        double b[], double se[], Integer *dropvar, 
                        Integer info[], FILE *fp)
{
  /* Select a covariate to drop from a Cox proportional hazards model using 
     the Wald statistic and its associated p-value 
  
     Input:
     m,pd and
     fp      - As per backward_selection
     csz     - On entry: indicates which subset of covariates are in the current
               model, with 
                csz[j+1] >= 1 indicating that the jth covariate is in the
                              current model 
                csz[j+1] = 0  indicating that the jth covariate is not in the 
                              current model 
               On exit: as per entry, but if dropvar >= 1, then csz[dropvar-1] 
               will have been set to 0 as well (i.e. on exit csz indicates 
               the covariates included in the new model, with covariate 
               indicated by dropvar having been dropped) 
  
     Output: 
     dropvar - The covariate to add to the model, If dropvar > 0 then 
               1 <= dropvar <= m. If dropvar < 1 then no covariates satisfied 
               the criteria defined by pd and so no covariate was dropped from
               the model 
     info    - As per backward_selection 
  */

  /* Integer scalar and array declarations */
  Integer j, mj, v;

  /* NAG structures and types */
  NagError fail;
  Nag_Boolean wacc;

  /* double scalar and array declarations */
  double df, min_w, p, w;

  /* initialise the error structure */
  INIT_FAIL(fail);

  if (fp)
    {
      /* display column headers for monitoring information */
      fprintf(fp,"   Covariate    Score\n");
      v = -1;
    }

  /* initialise the variable holding the minimum score (min_w), the 
     associated covariate (mj) and the variable to drop */
  min_w = -1.0;
  mj = 0;
  *dropvar = -1;

  /* get the maximum wald score and the associated covariate */
  for (j = 0; j < *ip; j++)
    {
      /* this method of calculating the Wald statistic is assuming that all of
         the covariates are continuous or binary (i.e. they have a single 
         degree freedom). If they do not then the wald_test_statistic routine 
         would need to be called and this routine would require additional 
         input */
      if (se[j]!=0) {
        w = (b[j]/se[j]);
        w*=w;

        if (w<min_w || min_w<0.0)
          {
            min_w = w;
            mj = j;
          }
      }
      else
        {
          /* division by zero due to a standard error of zero this should not
             occur in usual usage */
          info[0] = NE_BAD_PARAM;
          info[3] = NE_VAR_EQ_ZERO;
          return;
        }

      if (fp)
        {
          /* display the individual scores */

          /* translate j into the actual covariate ID */
          for (v++; v < m; v++)
            if (csz[v]>=1) break;

          fprintf(fp,"   %5"NAG_IFMT"   %10.4f\n",v+1,w);
        }
    }

  /* degrees of freedom. This assumes that the covariates are either continuous
     or binary (i.e. they have a single degree of freedom). If this is not the 
     case then the calculation of the degrees of freedom would need changing */
  df = 1.0;

  /* get the p-value for min_w */
  wacc = Nag_FALSE;
  p = g01ecc(Nag_UpperTail,min_w,df,&fail);
  if (fail.code!=NE_NOERROR)
    {
      info[2] = fail.code;
      if (fail.code!=NE_ALG_NOT_CONV)
        {
          info[0] = NE_BAD_PARAM;
          return;
        }
      wacc = Nag_TRUE;
      /* fail.code = NE_ALG_NOT_CONV is an accuracy warning, but the returned
         value should be sufficient so continue */
    }

  /* find the covariate ID associated with mj */
  for (v = j = 0; v < m; v++)
    {
      if (csz[v]>=1) 
        {
          if (j==mj) break;
          j++;
        }
    }

  if (fp)
    {
      /* print summary information for this iteration */
      fprintf(fp,"\n   Covariate %"NAG_IFMT" considered for removal\n", v + 1);
      fprintf(fp,"       Score   = %10.4f\n", min_w);
      fprintf(fp,"       p-value = %10.6f\n", p);
      if (wacc)
        {
          fprintf(fp,"          convergence failed in calculation of p-value\n");
          fprintf(fp,"          reported value should be a good approximation\n");
        }
      fprintf(fp,"       cut-off = %10.6f\n\n", pd);
    }

  /* check whether the p-value associated with the largest value of W meets */
  /* the criteria defined by pd */
  if (p<pd)
    {
      *dropvar = -1;
    }
  else
    {
      *dropvar = v + 1;

      /* drop the covariate associated with largest value of W from the model */
      csz[v] = 0;
      (*ip)--;
    }
}
